/*
Program to report on library dependencies.

Note that although the code and comments
talk about lists, STL maps are actually used.
This is for efficiency since lookup by
symbol name is a very common operation
and needs to be fast.
*/

#include <map>
#include <cstring>
#include <iostream>

#include <cstdlib>

#include <cstdio>

struct MSL; // fwd decl

// defined symbol list item

/*
 * A symbol_list is a list of pointers to MSL items.
 * An MSL item is a symbol name with an associated library.
 * The library is the one the symbol is defined in.
 * A library has two symbols lists.
 *
 * dsl - defined symbols list.
 *       This is the list of symbols defined in the library.
 * rsl - referenced symbols list.
 *       This is the list of symbols that this library refers to
 *       but are not defined in the library.
 */
class symbol_list
{
public:
  MSL* symbol_;

  symbol_list(MSL* msl_item) : symbol_(msl_item) {}
};

struct item_compare 
{
    bool operator() (const std::string & lhs, const std::string& rhs) const 
    {
        return lhs < rhs;
    }
};

struct LL;

typedef std::map<std::string, symbol_list*, item_compare> sl_type;
typedef std::map<std::string, symbol_list*, 
                 item_compare>::iterator sl_type_iterator;

typedef std::map<std::string, LL*, item_compare> ll_type;
typedef std::map<std::string, LL*, 
                 item_compare>::iterator ll_type_iterator;

/*
 * A library list is a list of libraries where each
 * item on the list contains a DSL and an RSL and
 * an RLL. The RLL is a list of libraries that this
 * library refers to. The RLL is empty initially
 * but can be built once the DSLs and RSLs are
 * complete for all libraries. The RLL is
 * essentially the adjacency list for a given
 * library which enables a directed acyclic graph
 * (DAG) to be produced.
 */
struct LL
{
public:
    std::string library_name_;

    sl_type dsl_;
    sl_type rsl_;
    ll_type rll_;
    int node_number;

    LL(const std::string& library_name)
        : library_name_(library_name)
    {
        node_number = next_node_number;
        next_node_number++;
    };

    static int next_node_number;
};

/*static*/ int LL::next_node_number = 1;

/*
 * Master symbol list item.
 * Symbols must be unique (if not then an error is
 * reported and the first occurence used).
 * Each symbol has an entry in the MSL.
 * The entry contains the symbol name and a pointer
 * to the library description for the library that
 * defines that symbol.
 */
class MSL
{
public:
    std::string symbol_name_;
    LL* library_;
    bool isDefined_;
    MSL(const std::string& symbol_name, LL* library, bool isDefined)
    : symbol_name_(symbol_name), library_(library), isDefined_(isDefined) {}
};

typedef std::map<std::string, MSL*, item_compare> msl_type;
typedef std::map<std::string, MSL*, item_compare>::iterator msl_type_iterator;

msl_type msl;

ll_type ll;

MSL* find_msl_symbol(const std::string& symbol)
{
    msl_type_iterator it = msl.find(symbol);
    if (it == msl.end())
        return 0;

    return (*it).second;
}

MSL* find_symbol_in_dsl(LL* lib_item, const std::string& symbol)
{
    sl_type_iterator it = lib_item->dsl_.find(symbol);
    if (it == lib_item->dsl_.end())
        return 0;
  return (*it).second->symbol_;
}

/*
  This adds the msl item if not there already.
  If there is an msl item there already and
  the library is defined then this is an error.
  it means that two libraries define the same symbol.
 */
MSL* addMSLitem(const std::string& symbol, LL* library, bool isDefined)
{
    msl_type_iterator it = msl.find(symbol);
    if (it != msl.end())
    {
        MSL* the_msl = (*it).second;
        if (!the_msl->isDefined_)
        {
            // this is not an error, it means
            // that we have found another symbol reference
            // for the same symbol.

            the_msl->library_ = library;
            the_msl->isDefined_ = isDefined;
            return the_msl;
        }

        if (!isDefined)
        {
            // this is not an error.
            // this happens when we find a reference
            // to a symbol and we have already seen
            // the definition.

            return the_msl;
        }

        if (library->library_name_ != the_msl->library_->library_name_)
        {
            std::cerr << "Error: Symbol " << symbol
                      << " is in " << library->library_name_
                      << " and " << the_msl->library_->library_name_
                      << std::endl;
        }

        return 0;
    }

    MSL* item = new MSL(symbol, library, isDefined);
    msl[symbol] = item;
    return item;
}

void add_defined_symbol(LL* lib_item, MSL* msl_item)
{
    sl_type_iterator it = lib_item->dsl_.find(msl_item->symbol_name_);
    if (it == lib_item->dsl_.end())
    {
        symbol_list* slist = new symbol_list(msl_item);
        lib_item->dsl_[msl_item->symbol_name_] = slist;
    }
}

void add_referenced_symbol(LL* lib_item, MSL* msl_item)
{
    sl_type_iterator it = lib_item->rsl_.find(msl_item->symbol_name_);
    if (it == lib_item->rsl_.end())
    {
        symbol_list* slist = new symbol_list(msl_item);
        lib_item->rsl_[msl_item->symbol_name_] = slist;
    }
}

LL* add_library(const std::string& library_name)
{
    LL* item = 0;
    ll_type_iterator it = ll.find(library_name);
    if (it == ll.end())
    {
        item = new LL(library_name);
        ll[library_name] = item;
    }
    else
    {
        item = (*it).second;
    }

   return item;
}

void add_library_to_rll(LL* lib_item, LL* referenced)
{
    ll_type_iterator it = lib_item->rll_.find(referenced->library_name_);
    if (it == lib_item->rll_.end())
    {
        lib_item->rll_[referenced->library_name_] = referenced;
    }
}

/*
 * This builds the adjacency list for each library.
 */
void build_rll(LL* lib_item)
{
    // for each referenced symbol in this library,
    // find which library it belongs to.

    sl_type_iterator it = lib_item->rsl_.begin();
    while (it != lib_item->rsl_.end())
    {
        symbol_list* the_symbol_list_item = (*it).second;
        MSL* msl_item = find_msl_symbol(the_symbol_list_item->symbol_->symbol_name_);
        if (msl_item != 0 && msl_item->library_ != 0)
        {
            if (msl_item->library_->library_name_ != lib_item->library_name_)
            {
                add_library_to_rll(lib_item, msl_item->library_);
            }
        }
        ++it;
    }
}

void add_symbol_info(const std::string& library_name,
                     const std::string symbol_name,
                     bool is_defined)
{
    LL* libitem = add_library(library_name);
    MSL* msl_item;

    // we need to add this as a symbol
    // even if we dont know which library defines it yet.

    if (is_defined)
    {
        msl_item = addMSLitem(symbol_name, libitem, true);
        if (msl_item != 0)
        {
            add_defined_symbol(libitem, msl_item);
        }
    }
    else
    {
        msl_item = addMSLitem(symbol_name, 0, false);
        if (msl_item != 0)
        {
            add_referenced_symbol(libitem, msl_item);
        }
    }
}

int main(int argc, char* argv[])
{
    char buffer[10240];
    char library_name[1024];
    char symbol_name [10240];
    int defined;
    char* tok;
    const char* delimiters = "|";
    FILE* fptr;

    if (argc == 1)
    {
        fptr = stdin;
    }
    else
    {
        fptr = fopen(argv[1], "r");
        if (fptr == 0)
        {
            perror("fopen");
            return EXIT_FAILURE;
        }
    }

    // Input records are of the form:
    // libname|symbol|defined
    // where <defined> is 1 for defined and 0 for referenced.

    while (fgets(buffer, sizeof(buffer), fptr))
    {
        buffer[strlen(buffer)-1] = '\0';
        tok = strtok(buffer, delimiters);
        if (tok == 0)
        {
            std::cerr << "Parse error on input." << std::endl;
            return 1;
        }
        strcpy(library_name, tok);
        tok = strtok(NULL, delimiters);
        if (tok == 0)
        {
            std::cerr << "Parse error on input." << std::endl;
            return 1;
        }
        strcpy(symbol_name, tok);
        tok = strtok(NULL, delimiters);
        if (tok == 0)
        {
            std::cerr << "Parse error on input." << std::endl;
            return 1;
        }
        defined = atoi(tok);

        add_symbol_info(library_name, symbol_name, defined);
    }

    if (argc != 1)
    {
        fclose(fptr);
    }

    // now we have defined all the data, we can
    // examine the lists to see which libraries
    // refer to which other libraries.

    ll_type_iterator it = ll.begin();
    while (it != ll.end())
    {
        build_rll((*it).second);
        ++it;
    }

    // print which library depends on which library

    std::cout << "digraph lib {" << std::endl;

    it = ll.begin();
    while (it != ll.end())
    {
        LL* item = (*it).second;
        std::cout << "node_" << item->node_number << " [label=\"";
        std::cout << item->library_name_ << "\"]\n";
        ++it;
    }

    it = ll.begin();
    while (it != ll.end())
    {
        LL* item = (*it).second;
        if (!(item->rll_.empty()))
        {
            ll_type_iterator it2 = item->rll_.begin();
            while (it2 != item->rll_.end())
            {
                LL* item2 = (*it2).second;
                std::cout << "node_" << item->node_number 
                          << "->" << "node_" << item2->node_number 
                          << std::endl;
                ++it2;
            }
        }
        ++it;
    }

    std::cout << "}\n";

    return EXIT_SUCCESS;
}
